import uuid
from datetime import datetime
from pathlib import Path
from collections import defaultdict

from pafts.datasets.dataset import Dataset

from silero_vad import load_silero_vad, read_audio, get_speech_timestamps
from pydub import AudioSegment
from pyannote.audio import Pipeline
import torch


def generate_unique_filename(format="wav"):
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    unique_id = uuid.uuid4().hex
    return f"temp_audio_{timestamp}_{unique_id}.{format}"


def diarization(
        dataset: Dataset,
        hf_token,

):
    """
    This function performs diarization (speaker separation) on the audio files,
    splits them by sentence, and then separates the audio by each speaker.

    Args:
        dataset (Dataset): Dataset instance.
        hf_token (str): Huggingface access token.

    Return:
        new_audios (list): List of new audio path.

    """

    pipeline = Pipeline.from_pretrained(
        "pyannote/speaker-diarization-3.1",
        use_auth_token=hf_token)

    # check using gpu or not
    if torch.cuda.is_available():
        device = torch.device("cuda")
        print("Using GPU...")
    else:
        device = torch.device("cpu")
        print("Using CPU...")

    pipeline.to(device)

    seg = None
    padding = AudioSegment.silent(duration=1000)

    for audio in dataset.audios:
        if not seg:
            seg = AudioSegment.from_file(audio)
        else:
            seg += AudioSegment.from_file(audio)
        seg += padding

    temp_file_path = dataset.output_path / generate_unique_filename()

    # create audio temp file
    seg.export(temp_file_path, format='wav')

    # diarization
    diarization_audio = pipeline(f"{temp_file_path}")

    speaker_num_list = defaultdict(int)
    new_audios = []

    for i, (turn, _, speaker) in enumerate(diarization_audio.itertracks(yield_label=True)):
        start_ms = turn.start * 1000
        end_ms = turn.end * 1000

        speaker_folder = dataset.output_path / f"speaker_{speaker}"
        speaker_folder.mkdir(parents=True, exist_ok=True)

        segment = seg[start_ms:end_ms]

        output_file_path = speaker_folder / f"{speaker}_{speaker_num_list[speaker]}.wav"
        segment.export(output_file_path, format="wav")

        speaker_num_list[speaker] += 1  # +1

        new_audios.append(speaker_folder / f"{speaker}_{speaker_num_list[speaker]}.wav")

    # delete temp file
    if Path(temp_file_path).exists():
        temp_file_path.unlink()

    dataset.audios = new_audios

    return new_audios


def vad(
        dataset: Dataset,
        min_silence_duration_ms=500,
        padding_duration_ms=200,

):
    model = load_silero_vad()
    audios = dataset.audios
    stamp = []
    new_audios = []

    for audio in audios:
        wav = read_audio(str(audio))

        speech_timestamps = get_speech_timestamps(wav, model, min_silence_duration_ms=min_silence_duration_ms)

        for speech_timestamp in speech_timestamps:
            start = speech_timestamp['start']
            end = speech_timestamp['end']
            start /= 16000
            end /= 16000
            stamp.append((start, end))

        audio_segment = AudioSegment.from_file(audio)

        for i, (start, end) in enumerate(stamp):
            start_ms = start * 1000
            end_ms = end * 1000

            segment = audio_segment[start_ms:end_ms]
            padding = AudioSegment.silent(duration=padding_duration_ms)
            segment = padding + segment + padding

            file_name = audio.with_name(f'{audio.stem}_{i}{audio.suffix}')
            file_name = file_name.name

            path = (dataset.output_path / file_name).resolve()
            segment.export(path, format=audio.suffix[1:])

            new_audios.append(path)

    dataset.audios = new_audios

    return new_audios



import json
from pathlib import Path

import whisper
from whisper.tokenizer import LANGUAGES
from tqdm import tqdm

from pafts.datasets.dataset import Dataset

whisper_model = {
    "tiny": None,
    "base": None,
    "small": None,
    "medium": None,
    "large": None
}


def whisper_stt(
        audio: Path,
        model_size='base',
        language=None,
):
    """
        Use the Whipser[https://github.com/openai/whisper] STT model to extract text.
        If there is no gpu or low performance, use the base model.

        Args:
            data (Data): Audio data.
            model_size (str): Size of the whisper model.
            language (str): Language of the audio file to run STT.

        Return:
            str: Text in an audio file.

        """
    global whisper_model
    if model_size not in whisper_model.keys():
        raise ValueError(
            "[!] Invalid model size selected. Please choose one of the following sizes: tiny, base, small, medium, large.")

    if not whisper_model[model_size]:
        whisper_model[model_size] = whisper.load_model(model_size)

    if language and language not in LANGUAGES:
        raise ValueError(
            f"[!] This language is not supported. Please select one of the language codes below\n{LANGUAGES}")

    result = whisper_model[model_size].transcribe(str(audio))

    return result['text']


def STT(
        dataset: Dataset,
        output_format='json',
        model_size='base',
        language=None
):
    """
    Read the audio files in the dataset, and use the stt function to extract text.
    Save the extracted text in the form of a json file.

    Args:
        dataset (Dataset): Audio dataset Class
        output_path (str): Output path
        format (str): Output format, Defaults is json (json or txt)
        model_size (str): Size of the whisper model.
        language (str): Language of the audio file to run STT.

    Returns:
        Dict: Dictionary of the text values of audio files in the dataset.

    """
    audios = dataset.audios
    stt_dict = {}

    output_path = dataset.output_path

    bar = tqdm(audios,
               total=len(audios),
               leave=True,
               )

    for audio in bar:
        text = whisper_stt(audio, model_size, language)
        stt_dict[audio.name] = text

    if output_format == 'json':
        with open(Path(output_path) / Path(f'{dataset.dataset_name}.json'), 'w', encoding='UTF8') as f:
            json.dump(stt_dict, f, indent=4, ensure_ascii=False)
    elif output_format == 'txt':
        with open(Path(output_path) / Path(f'{dataset.dataset_name}.txt'), 'w', encoding='UTF8') as f:
            for k, v in stt_dict:
                f.write(f'{k}|{v}\n')
    else:
        raise ValueError(
            f"[!] Please choose one of the following format: json, txt.")

    return stt_dict



import shutil
from pathlib import Path
import uuid
from datetime import datetime

from pafts.datasets.dataset import Dataset
from pafts.diarization.diarization import diarization
from pafts.separator.separator import separator
from pafts.stt.stt import STT


class PAFTS:
    """
        Make audio files into a dataset for TTS.

        Args:
        path (str): Directory path with audio files.
        dataset_name (str, optional): Dataset name. Defaults to dataset_path's directory name.
        language (str, optional): Language using BCP 47 language tag. Defaults to 'en-us' (English)
        output_path (str): Output Directory. Defaults to './pafts_output'

        Example with quick start:
        p = PAFTS(
            path = 'your_audio_directory_path',
            output_path = 'output_path',
            hf_token="HUGGINGFACE_ACCESS_TOKEN_GOES_HERE"
        )
        p.run()


        """

    def __init__(
            self,
            path: str = None,
            dataset_name: str = None,
            language: str = None,
            output_path: str = 'pafts_output',
            hf_token: str = None
    ):

        self._hf_token = hf_token

        self._dataset = Dataset(
            path=path,
            dataset_name=dataset_name,
            language=language,
            output_path=output_path
        )

    def separator(self):
        separator(self._dataset)
        return

    def diarization(self):
        if not self._hf_token:
            raise TypeError("[!] Hugging Face access token is required to use diarization model.")

        diarization(self._dataset, self._hf_token)

        return

    def stt(self, output_format='json', model_size='large'):
        STT(self._dataset, output_format=output_format, model_size=model_size)
        return

    def _stage_process(self, process_function, *args, **kwargs):
        # Create unique temp directory
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        unique_id = uuid.uuid4().hex
        temp_dir = Path.cwd() / f"temp_dir_{timestamp}_{unique_id}"
        temp_dir.mkdir(exist_ok=True)

        self._dataset.output_path = temp_dir
        process_function(self._dataset, *args, **kwargs)

        # Update dataset path
        self._dataset.path = temp_dir

        return temp_dir

    def run(self):
        original_output = self._dataset.output_path

        # Stage 1: separator
        temp_dir1 = self._stage_process(separator)

        # Stage 2: diarization
        temp_dir2 = self._stage_process(diarization, hf_token=self._hf_token)

        # Stage 3: STT
        temp_dir3 = self._stage_process(STT)

        original_output.mkdir(exist_ok=True)

        for temp_dir in [temp_dir2, temp_dir3]:
            for file in temp_dir.rglob('*'):
                if file.is_file():
                    relative_path = file.relative_to(temp_dir)
                    destination = original_output / relative_path
                    destination.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(file, destination)

        # Cleanup
        for temp_dir in [temp_dir1, temp_dir2, temp_dir3]:
            shutil.rmtree(temp_dir, ignore_errors=True)
